#include "bilInternal.h"


static BILboolean clipToRectangle(BILint  srcX1, BILint  srcY1, BILsizei  srcWidth, BILsizei  srcHeight,
                                  BILint  dstX1, BILint  dstY1, BILsizei  dstWidth, BILsizei  dstHeight,
                                  BILint *outX1, BILint *outY1, BILsizei *outWidth, BILsizei *outHeight)
  {
  BILint srcX2, srcY2;
  BILint dstX2, dstY2;

  bilCheckParameter(outX1     != NULL, return BIL_FALSE);
  bilCheckParameter(outY1     != NULL, return BIL_FALSE);
  bilCheckParameter(outWidth  != NULL, return BIL_FALSE);  
  bilCheckParameter(outHeight != NULL, return BIL_FALSE);

  /* too small to be visible? */
  if ((srcWidth <= 0) || (srcHeight <= 0) ||
      (dstWidth <= 0) || (dstHeight <= 0))
    return BIL_FALSE;

  /* calculate extents */
  srcX2 = srcX1 + srcWidth  - 1;
  srcY2 = srcY1 + srcHeight - 1;
  dstX2 = dstX1 + dstWidth  - 1;
  dstY2 = dstY1 + dstHeight - 1;

  /* no overlap? */
  if ((srcX2 < dstX1) || (srcX1 > dstX2) ||
      (srcY2 < dstY1) || (srcY1 > dstY2))
    return BIL_FALSE;

  /* clip x */
  if (srcX1 < dstX1) srcX1 = dstX1;
  if (srcX2 > dstX2) srcX2 = dstX2;

  /* clip y */
  if (srcY1 < dstY1) srcY1 = dstY1;
  if (srcY2 > dstY2) srcY2 = dstY2;

  /* return clipped rectangle */
  *outX1     = srcX1;    
  *outY1     = srcY1;
  *outWidth  = srcX2 - srcX1 + 1;
  *outHeight = srcY2 - srcY1 + 1;

  /* non-overlap should have been caught already */
  assert(*outWidth  > 0);
  assert(*outHeight > 0);

  return BIL_TRUE;
  }


BILboolean bilDrawPixels(BILimage _image, BILint x, BILint y, BILsizei width, BILsizei height, BILformat format, BILtype type, const BILvoid *pixels)
  {
  BILimageStruct     *image = _image;  
  BILconvertFunction  convertScanline;
  BILint              clippedX;
  BILint              clippedY;
  BILsizei            clippedWidth;
  BILsizei            clippedHeight;
  BILsizei            srcBytesPerPixel, srcBytesPerScanline;
  BILsizei            dstBytesPerPixel, dstBytesPerScanline;
  const BILubyte     *srcPixels;
  BILubyte           *dstPixels;

  bilCheckParameter(image  != NULL, return BIL_FALSE);
  bilCheckParameter(pixels != NULL, return BIL_FALSE);

  /* find conversion function */
  convertScanline = bilGetConvertFunction(format, type, BIL_RGBA, BIL_UNSIGNED_BYTE);
  if (!convertScanline)
    return BIL_FALSE;

  /* clip input coordinates */
  if (!clipToRectangle(x,         y,         width,         height,
                       0,         0,         image->width,  image->height,
                       &clippedX, &clippedY, &clippedWidth, &clippedHeight))
    return BIL_TRUE;

  /* make sure our clipping worked ok (so we can use fast loop below) */
  assert(clippedHeight > 0);
  assert(clippedWidth  > 0);

  /* setup src */
  srcBytesPerPixel    = bilBytesPerPixel(format, type);
  srcBytesPerScanline = srcBytesPerPixel * width;
  srcPixels           = (BILubyte*)pixels + ((clippedY-y) * srcBytesPerScanline) + ((clippedX-x) * srcBytesPerPixel);

  /* setup dst */
  dstBytesPerPixel    = sizeof(BILpixelStruct);
  dstBytesPerScanline = dstBytesPerPixel * image->width;
  dstPixels           = (BILubyte*)image->pixels + (clippedY * dstBytesPerScanline) + (clippedX * dstBytesPerPixel);

  /* draw pixels */
  do
    {
    convertScanline(dstPixels, srcPixels, clippedWidth);    
    srcPixels += srcBytesPerScanline;
    dstPixels += dstBytesPerScanline;
    } while (--clippedHeight != 0);

  return BIL_TRUE;
  }


BILboolean bilReadPixels(BILimage _image, BILint x, BILint y, BILsizei width, BILsizei height, BILformat format, BILtype type, BILvoid *pixels)
  {
  BILimageStruct     *image = _image;
  BILconvertFunction  convertScanline;
  BILint              clippedX;
  BILint              clippedY;
  BILsizei            clippedWidth;
  BILsizei            clippedHeight;
  BILsizei            srcBytesPerPixel, srcBytesPerScanline;
  BILsizei            dstBytesPerPixel, dstBytesPerScanline;
  const BILubyte     *srcPixels;
  BILubyte           *dstPixels;

  bilCheckParameter(image  != NULL, return BIL_FALSE);
  bilCheckParameter(pixels != NULL, return BIL_FALSE);

  /* find conversion function */
  convertScanline = bilGetConvertFunction(BIL_RGBA, BIL_UNSIGNED_BYTE, format, type);
  if (!convertScanline)
    return BIL_FALSE;

  /* clear pixels */
  dstBytesPerPixel = bilBytesPerPixel(format, type);
  memset(pixels, 0, width * height * dstBytesPerPixel);

  /* clip coordinates to image */
  if (!clipToRectangle(x,         y,         width,         height,
                       0,         0,         image->width,  image->height,
                       &clippedX, &clippedY, &clippedWidth, &clippedHeight))
    return BIL_TRUE;

  /* make sure our clipping worked ok (so we can use fast loop below) */
  assert(clippedHeight > 0);
  assert(clippedWidth  > 0);

  /* setup src */
  srcBytesPerPixel    = sizeof(BILpixelStruct);
  srcBytesPerScanline = srcBytesPerPixel * image->width;
  srcPixels           = (BILubyte*)image->pixels + (clippedY * srcBytesPerScanline) + (clippedX * srcBytesPerPixel);

  /* setup dst */  
  dstBytesPerScanline = dstBytesPerPixel * width;
  dstPixels           = (BILubyte*)pixels + ((clippedY-y) * dstBytesPerScanline) + ((clippedX-x) * dstBytesPerPixel);

  /* read pixels */  
  do
    {
    convertScanline(dstPixels, srcPixels, clippedWidth);
    srcPixels += srcBytesPerScanline;
    dstPixels += dstBytesPerScanline;
    } while (--clippedHeight != 0);

  return BIL_TRUE;
  }


BILboolean bilCopyPixels(BILimage _srcImage, BILint srcX, BILint srcY, BILsizei width, BILsizei height, BILimage _dstImage, BILint dstX, BILint dstY)
  {
  BILimageStruct     *srcImage = _srcImage;
  BILimageStruct     *dstImage = _dstImage;
  BILint              clippedSrcX,  clippedSrcY;
  BILint              clippedDstX,  clippedDstY;
  BILsizei            clippedWidth, clippedHeight;
  BILsizei            srcBytesPerScanline;
  BILsizei            dstBytesPerScanline;
  const BILubyte     *srcPixels;
  BILubyte           *dstPixels;

  bilCheckParameter(srcImage != NULL, return BIL_FALSE);
  bilCheckParameter(dstImage != NULL, return BIL_FALSE);

  /* clip coordinates to srcImage */
  if (!clipToRectangle(srcX,         srcY,         width,           height,
                       0,            0,            srcImage->width, srcImage->height,
                       &clippedSrcX, &clippedSrcY, &clippedWidth,   &clippedHeight))
    return BIL_TRUE;

  /* clip coordinates to dstImage */
  if (!clipToRectangle((clippedSrcX-srcX)+dstX, (clippedSrcY-srcY)+dstY, clippedWidth,     clippedHeight,
                       0,                       0,                       dstImage->width,  dstImage->height,
                       &clippedDstX,            &clippedDstY,            &clippedWidth,    &clippedHeight))
    return BIL_TRUE;

  /* make sure our clipping worked ok (so we can use fast loop below) */
  assert(clippedHeight > 0);
  assert(clippedWidth  > 0);

  /* setup src */
  srcBytesPerScanline = srcImage->width * sizeof(BILpixelStruct);
  srcPixels           = bilGetImagePtr(srcImage, clippedSrcX, clippedSrcY);

  /* setup dst */  
  dstBytesPerScanline = dstImage->width * sizeof(BILpixelStruct);
  dstPixels           = bilGetImagePtr(dstImage, clippedDstX, clippedDstY);

  /* copy pixels */  
  do
    {
    memmove(dstPixels, srcPixels, clippedWidth * sizeof(BILpixelStruct));
    srcPixels += srcBytesPerScanline;
    dstPixels += dstBytesPerScanline;
    } while (--clippedHeight != 0);

  return BIL_TRUE;
  }


