#include "bilInternal.h"

static BILmap gEncoders = { 0 };
static BILmap gDecoders = { 0 };
static BILmap gFilters  = { 0 };


static void _bilCleanDriverMaps(void)
  {
  bilMapDone(&gEncoders);
  bilMapDone(&gDecoders);
  bilMapDone(&gFilters);
  }


static void bilInitDriverMaps(void)
  {
  static BILboolean initialised = BIL_FALSE;

  if (!initialised)
    {
    bilMapInit(&gEncoders);
    bilMapInit(&gDecoders);
	bilMapInit(&gFilters);

    atexit(_bilCleanDriverMaps);
    initialised = BIL_TRUE;
    }
  }


static BILboolean bilRegisterDriver(BILmap        *map,
                                    const BILchar *name,
                                    const BILvoid *value,
                                    BILsizei       size)
  {
  bilCheckParameter(map   != NULL, return BIL_FALSE);
  bilCheckParameter(name  != NULL, return BIL_FALSE);
  bilCheckParameter(value != NULL, return BIL_FALSE);
  bilCheckParameter(size  >= 0,    return BIL_FALSE);

  /* init driver maps */
  bilInitDriverMaps();

  /* name must exist */
  if (strlen(name) <= 0)
    {
    bilSetError(BIL_INVALID_VALUE);
    return BIL_FALSE;
    }

  /* name cannot contain a space */
  if (strchr(name, ' '))
    {
    bilSetError(BIL_INVALID_VALUE);
    return BIL_FALSE;
    }

  /* encoder must not already exist */
  if (bilMapGetValue(map, name))
    return BIL_TRUE;

  return bilMapSetValue(map, name, value, size);
  }


static BILboolean bilUnregisterDriver(BILmap        *map,
                                      const BILchar *name)
  {
  bilCheckParameter(map  != NULL, return BIL_FALSE);
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return bilMapRemoveValue(&gEncoders, name);
  }


const BILvoid *bilGetDriver(const BILmap  *map,
                            const BILchar *name)
  {
  bilCheckParameter(map  != NULL, return BIL_FALSE);
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return bilMapGetValue(map, name);
  }


static BILboolean bilEnumerateDrivers(BILmap  *map,
                                      BILchar *string,
                                      BILsizei stringSize)
  {
  bilCheckParameter(map        != NULL, return BIL_FALSE);
  bilCheckParameter(string     != NULL, return BIL_FALSE);
  bilCheckParameter(stringSize >= 0,    return BIL_FALSE);

  return bilMapEnumarateKeys(map, string, stringSize);
  }


BILboolean bilRegisterEncoder(const BILchar *name, const BILencoderStruct *encoder)
  {
  bilCheckParameter(name                         != NULL, return BIL_FALSE);
  bilCheckParameter(encoder                      != NULL, return BIL_FALSE);
  bilCheckParameter(encoder->newEncodeContext    != NULL, return BIL_FALSE);
  bilCheckParameter(encoder->deleteEncodeContext != NULL, return BIL_FALSE);
  bilCheckParameter(encoder->beginEncode         != NULL, return BIL_FALSE);
  bilCheckParameter(encoder->endEncode           != NULL, return BIL_FALSE);
  bilCheckParameter(encoder->encodeBlock         != NULL, return BIL_FALSE);

  return bilRegisterDriver(&gEncoders, name, encoder, sizeof(BILencoderStruct));
  }


BILboolean bilUnregisterEncoder(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return bilUnregisterDriver(&gEncoders, name);
  }


const BILencoderStruct *bilGetEncoder(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return (BILencoderStruct *)bilGetDriver(&gEncoders, name);
  }


BILboolean bilEnumerateEncoders(BILchar *string, BILsizei stringSize)
  {
  return bilEnumerateDrivers(&gEncoders, string, sizeof(string));
  }


BILboolean bilRegisterDecoder(const BILchar *name, const BILdecoderStruct *decoder)
  {
  bilCheckParameter(name                         != NULL, return BIL_FALSE);
  bilCheckParameter(decoder                      != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->newDecodeContext    != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->deleteDecodeContext != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->beginDecode         != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->endDecode           != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->decodeBlock         != NULL, return BIL_FALSE);
  bilCheckParameter(decoder->validBlock          != NULL, return BIL_FALSE);

  return bilRegisterDriver(&gDecoders, name, decoder, sizeof(BILdecoderStruct));
  }


BILboolean bilUnregisterDecoder(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return bilUnregisterDriver(&gDecoders, name);
  }


const BILdecoderStruct *bilGetDecoder(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return (BILdecoderStruct *)bilGetDriver(&gDecoders, name);
  }


BILboolean bilEnumerateDecoders(BILchar *string, BILsizei stringSize)
  {
  return bilEnumerateDrivers(&gDecoders, string, sizeof(string));
  }


BILboolean bilRegisterFilter(const BILchar *name, const BILfilterStruct *filter)
  {
  bilCheckParameter(name             != NULL, return BIL_FALSE);
  bilCheckParameter(filter           != NULL, return BIL_FALSE);
  bilCheckParameter(filter->function != NULL, return BIL_FALSE);

  return bilRegisterDriver(&gFilters, name, filter, sizeof(BILfilterStruct));
  }


BILboolean bilUnregisterFilter(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return bilUnregisterDriver(&gFilters, name);
  }


const BILfilterStruct *bilGetFilter(const BILchar *name)
  {
  bilCheckParameter(name != NULL, return BIL_FALSE);

  return (BILfilterStruct *)bilGetDriver(&gDecoders, name);
  }


BILboolean bilEnumerateFilters(BILchar *string, BILsizei stringSize)
  {
  return bilEnumerateDrivers(&gFilters, string, sizeof(string));
  }