#include "bilInternal.h"
#include <math.h>

#ifndef M_PI
#  define M_PI 3.14159265358979323846264338
#endif

typedef struct
	{
    double p0, p2, p3;
    double q0, q1, q2, q3;
	} BILmitchellStruct;

typedef struct
	{
    double a;
    double i0a;
	} BILkaiserStruct;


static double bessel_i0(double x)
	{
	double epsilon = 1e-7;
    double sum, y, t;
	int i;
	
    sum = 1;
    y   = x*x/4;
    t   = y;
	
    for (i = 2; t > epsilon; i++)
		{
		sum += t;
		t   *= (double)y/(i*i);
		}
	
    return sum;
	}


static BILdouble _bilBoxFunction(BILdouble x, BILvoid *d)
	{
    if (x < -0.5) return 0;
    if (x <  0.5) return 1;
    return 0;
	}


static BILdouble _bilTriangleFunction(BILdouble x, BILvoid *d)
	{
    if (x < -1) return 0;
    if (x <  0) return 1+x;
    if (x <  1) return 1-x;
    return 0;
	}


static BILdouble _bilQuadraticFunction(BILdouble x, BILvoid *d)
	{
    BILdouble t;
	
    if (x < -1.5) return 0;
    if (x < -0.5) {t = x+1.5; return 0.5*t*t;}
    if (x <  0.5) return 0.75-x*x;
    if (x <  1.5) {t = x-1.5; return 0.5*t*t;}
    return 0;
	}


static BILdouble _bilCubicFunction(BILdouble x, BILvoid *d)
	{
    BILdouble t;
	
    if (x < -2) return 0;
    if (x < -1) {t = 2+x; return t*t*t/6;}
    if (x <  0) return (4+x*x*(-6+x*-3))/6;
    if (x <  1) return (4+x*x*(-6+x*3))/6;
    if (x <  2) {t = 2-x; return t*t*t/6;}
    return 0;
	}


static BILdouble _bilCatromFunction(BILdouble x, BILvoid *d)
	{
    if (x < -2) return 0;
    if (x < -1) return 0.5*(4+x*(8.+x*(5+x)));
    if (x <  0) return 0.5*(2+x*x*(-5+x*-3));
    if (x <  1) return 0.5*(2+x*x*(-5+x*3));
    if (x <  2) return 0.5*(4+x*(-8+x*(5-x)));
    return 0;
	}


static BILdouble _bilGaussianFunction(BILdouble x, BILvoid *d)
	{
    return exp(-2*x*x)*sqrt(2/M_PI);
	}


static BILdouble _bilSincFunction(BILdouble x, BILvoid *d)
	{
    return (x==0) ? 1 : sin(M_PI*x)/(M_PI*x);
	}


static BILdouble _bilBesselFunction(BILdouble x, BILvoid *d)
	{
    /*
     * See Pratt "Digital Image Processing" p. 97 for Bessel functions
     * zeros are at approx x=1.2197, 2.2331, 3.2383, 4.2411, 5.2428, 6.2439,
     * 7.2448, 8.2454
     */
    return (x==0) ? M_PI/4 : j1(M_PI*x)/(2*x);
	}


static BILdouble _bilMitchellFunction(BILdouble x, BILvoid *d)
	{
    BILmitchellStruct *m = (BILmitchellStruct *)d;
	
    /*
     * see Mitchell&Netravali, "Reconstruction Filters in Computer Graphics",
     * SIGGRAPH 88
     */
    if (x < -2) return 0;
    if (x < -1) return m->q0-x*(m->q1-x*(m->q2-x*m->q3));
    if (x <  0) return m->p0+x*x*(m->p2-x*m->p3);
    if (x <  1) return m->p0+x*x*(m->p2+x*m->p3);
    if (x <  2) return m->q0+x*(m->q1+x*(m->q2+x*m->q3));
    return 0;
	}


static BILvoid _bilMitchellInit(BILdouble b, BILdouble c, BILvoid *d)
	{
    BILmitchellStruct *m = (BILmitchellStruct *)d;
	
    m->p0 = (  6 -  2*b       ) / 6;
    m->p2 = (-18 + 12*b +  6*c) / 6;
    m->p3 = ( 12 -  9*b -  6*c) / 6;
    m->q0 = (	    8*b + 24*c) / 6;
    m->q1 = (	 - 12*b - 48*c) / 6;
    m->q2 = (	    6*b + 30*c) / 6;
    m->q3 = (    -    b -  6*c) / 6;
	}


static BILdouble _bilHanningFunction(BILdouble x, BILvoid *d)
	{
    return 0.5+0.5*cos(M_PI*x);
	}


static BILdouble _bilHammingFunction(BILdouble x, BILvoid *d)
	{
    return 0.54+0.46*cos(M_PI*x);
	}


static BILdouble _bilBlackmanFunction(BILdouble x, BILvoid *d)
	{
    return 0.42+0.50*cos(M_PI*x)+0.08*cos(2*M_PI*x);
	}


static BILdouble _bilKaiserFunction(BILdouble x, BILvoid *d)
	{
    BILkaiserStruct *k = (BILkaiserStruct *)d;
	
	return bessel_i0(k->a*sqrt(1-x*x))*k->i0a;
	}


static BILvoid _bilKaiserInit(BILdouble a, BILdouble b, BILvoid *d)
	{
	BILkaiserStruct *k = (BILkaiserStruct *)d;
    
    k->a   = a;
    k->i0a = 1/bessel_i0(a);
	}


BILboolean bilPointRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter = 
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 0.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilBoxFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("point", &filter);
	}


BILboolean bilBoxRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter = 
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 0.5,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilBoxFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("box", &filter);
	}


BILboolean bilTriangleRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter = 
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilTriangleFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("triangle", &filter);
	}


BILboolean bilQuadraticRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter = 
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.5,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilQuadraticFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("quadratic", &filter);
	}


BILboolean bilCubicRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter = 
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 2.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilCubicFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("cubic", &filter);
	}


BILboolean bilCatromRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 2.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 0.0,
		/* data                */ NULL,
		/* function            */ _bilCatromFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("catrom", &filter);
	}


BILboolean bilMitchellRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILmitchellStruct data;

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 2.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 0.0,
		/* unitrange           */ 0.0,
		/* data                */ &data,
		/* function            */ _bilMitchellFunction,
		/* init                */ _bilMitchellInit
        };
	
	return bilRegisterFilter("mitchell", &filter);
	}


BILboolean bilGaussianRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.25,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 0.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilGaussianFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("gaussian", &filter);
	}


BILboolean bilSincRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 4.0,
		/* blur                */ 1.0,
		/* window              */ 1.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 0.0,
		/* data                */ NULL,
		/* function            */ _bilSincFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("sinc", &filter);
	}


BILboolean bilBesselRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 3.2383,
		/* blur                */ 1.0,
		/* window              */ 1.0,
		/* cardinal            */ 0.0,
		/* unitrange           */ 0.0,
		/* data                */ NULL,
		/* function            */ _bilBesselFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("bessel", &filter);
	}


BILboolean bilHanningRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilHanningFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("hanning", &filter);
	}


BILboolean bilHammingRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilHammingFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("hamming", &filter);
	}


BILboolean bilBlackmanRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };

	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ NULL,
		/* function            */ _bilBlackmanFunction,
		/* init                */ NULL
        };
	
	return bilRegisterFilter("blackman", &filter);
	}


BILboolean bilKaiserRegisterFilter()
	{
	static BILparameterStruct parameters[] =
		{
        { NULL, 0, 0 }
        };
	
	static BILkaiserStruct data;
	
	static BILfilterStruct filter =
       	{
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BIL_VERSION,
        /* parameters          */ parameters,
		/* supp                */ 1.0,
		/* blur                */ 1.0,
		/* window              */ 0.0,
		/* cardinal            */ 1.0,
		/* unitrange           */ 1.0,
		/* data                */ &data,
		/* function            */ _bilKaiserFunction,
		/* init                */ _bilKaiserInit
        };
	
	return bilRegisterFilter("kaiser", &filter);
	}


BILboolean bilRegisterFilters()
	{
	return bilPointRegisterFilter()     &&
		   bilBoxRegisterFilter()       &&
		   bilTriangleRegisterFilter()  &&
		   bilQuadraticRegisterFilter() &&
		   bilCubicRegisterFilter()     &&
		   bilCatromRegisterFilter()    &&
		   bilMitchellRegisterFilter()  &&
		   bilGaussianRegisterFilter()  &&
		   bilSincRegisterFilter()      &&
		   bilBesselRegisterFilter()    &&
		   bilHanningRegisterFilter()   &&
		   bilHammingRegisterFilter()   &&
		   bilBlackmanRegisterFilter()  &&
		   bilKaiserRegisterFilter();
	}
