#include "bilInternal.h"


BILimage bilNewImage(BILsizei width, BILsizei height)
  {
  BILimageStruct *image;

  bilCheckParameter(width  >= 0, return NULL);
  bilCheckParameter(height >= 0, return NULL);

  image = bilMalloc(sizeof(BILimageStruct));
  if (!image)
    return NULL;

  image->signature = BIL_IMAGE_SIGNATURE;
  image->width     = width;
  image->height    = height;
  image->pixels    = NULL;

  if (!bilMapInit(&image->parasites))
    {
    bilFree(image);
    return NULL;
    }

  image->pixels = bilMalloc(width * height * sizeof(BILpixelStruct));
  if (!image->pixels)
    {
    bilFree(image);
    return NULL;
    }

  bilClearImage(image);

  return (BILimage)image;
  }


BILimage bilCopyImage(BILimage source)
  {
  BILimage   image;
  BILsizei   width;
  BILsizei   height;
  BILboolean ok;

  bilCheckParameter(source != NULL, return NULL);

  width  = bilGetImageWidth(source);
  height = bilGetImageHeight(source);

  image = bilNewImage(width, height);
  if (!image)
    ok = BIL_FALSE;

  if (ok && !bilCopyPixels(source, 0, 0, width, height, image, 0, 0))
    ok = BIL_FALSE;

  if (ok && !bilCopyImageParasites(source, image))
    ok = BIL_FALSE;

  if (!ok)
    {
    if (image) bilDeleteImage(image);
    return NULL;
    }

  return image;
  }


BILboolean bilDeleteImage(BILimage _image)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image            != NULL,                return BIL_FALSE);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return BIL_FALSE);

  /* clear signature (for debugging) */
  image->signature = 0;

  /* release memory */
  bilMapDone(&image->parasites);
  bilFree(image->pixels);
  bilFree(image);

  return BIL_TRUE;
  }


BILboolean bilResizeImage(BILimage _image, BILsizei width, BILsizei height)
  {
  BILimageStruct *image = _image;
  BILpixelStruct *newPixels;

  bilCheckParameter(image            != NULL,                return BIL_FALSE);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return BIL_FALSE);
  bilCheckParameter(width            >= 0,                   return BIL_FALSE);
  bilCheckParameter(height           >= 0,                   return BIL_FALSE);

  if ((width != image->width) || (height != image->height))
    {
    newPixels = bilRealloc(image->pixels, width * height * sizeof(BILpixelStruct));
    if (!newPixels)
      return BIL_FALSE;

    image->pixels = newPixels;
    image->width  = width;
    image->height = height;
    }

  return BIL_TRUE;
  }


BILboolean bilClearImage(BILimage _image)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image            != NULL,                return BIL_FALSE);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return BIL_FALSE);

  memset(image->pixels, 0, image->width * image->height * sizeof(BILpixelStruct));

  return BIL_TRUE;
  }


BILvoid *bilGetImagePtr(BILimage _image, BILint x, BILint y)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image            != NULL,                return NULL);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return NULL);

  if ((x >= 0)           && (y >= 0) &&
      (x < image->width) && (y <  image->height))
    return image->pixels + (y * image->width) + x;
  else
    return NULL;
  }


BILsizei bilGetImageWidth(BILimage _image)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image            != NULL,                return 0);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return 0);

  return image->width;
  }


BILsizei bilGetImageHeight(BILimage _image)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image            != NULL,                return 0);
  bilCheckParameter(image->signature == BIL_IMAGE_SIGNATURE, return 0);

  return image->height;
  }


BILsizei bilBytesPerPixel(BILformat format, BILtype type)
  {
  BILsizei numComponents;
  BILsizei bytesPerComponent;
  
  switch (format)
    {
    case BIL_RGB:             numComponents = 3; break;
    case BIL_RGBA:            numComponents = 4; break;
    case BIL_RED:             numComponents = 1; break;
    case BIL_GREEN:           numComponents = 1; break;
    case BIL_BLUE:            numComponents = 1; break;
    case BIL_ALPHA:           numComponents = 1; break;
    case BIL_LUMINANCE:       numComponents = 1; break;
    case BIL_LUMINANCE_ALPHA: numComponents = 2; break;
    default:                  numComponents = 0;
                              bilAssert(0, bilSetError(BIL_INVALID_ENUM));
    }

  switch (type)
    {
    case BIL_UNSIGNED_BYTE:  bytesPerComponent = sizeof(BILubyte);  break;
    case BIL_UNSIGNED_SHORT: bytesPerComponent = sizeof(BILushort); break;
    case BIL_FLOAT:          bytesPerComponent = sizeof(BILfloat);  break;
    default:                 bytesPerComponent = 0;
                             bilAssert(0, bilSetError(BIL_INVALID_ENUM));
    }

  return numComponents * bytesPerComponent;
  }

