#include "bilInternal.h"


#define BIL_MAP_GRANULARITY	(64)


static BILboolean bilMapSetCapacity(BILmap    *map,
                                    BILsizei   newCapacity,
                                    BILboolean force)
  {
  bilCheckParameter(map           != NULL,        return BIL_FALSE);
  bilCheckParameter(map->length   >= 0,           return BIL_FALSE);
  bilCheckParameter(map->capacity >= map->length, return BIL_FALSE);

  if (!force)
    {
    /* dont shrink memory block */
    if (newCapacity < map->capacity)
      newCapacity = map->capacity;

    /* round capacity to granularity */
    newCapacity += BIL_MAP_GRANULARITY - 1;
    newCapacity /= BIL_MAP_GRANULARITY;
    newCapacity *= BIL_MAP_GRANULARITY;
    }

  /* resize block */
  if (map->capacity != newCapacity)
    {
    BILmapItem *newItems;

    newItems = bilRealloc(map->items, newCapacity * sizeof(BILmapItem));
    if (!newItems)
      {
      bilSetError(BIL_OUT_OF_MEMORY);
      return BIL_FALSE;
      }

    map->items    = newItems;
    map->capacity = newCapacity;
    }

  return BIL_TRUE;
  }


static BILboolean bilMapFind(const BILmap  *map,
                             const BILchar *key,
                             BILint        *index)
  {
  BILint     low, high, mid;
  BILint     cmp;
  BILboolean found;

  found = BIL_FALSE;
  low   = 0;
  high  = map->length;

  while (low < high)
    {
    mid = (low + high) / 2;
    cmp = strcmp(key, map->items[mid].key);

    if (cmp < 0)
      high = mid;
    else if (cmp > 0)
      low  = mid+1;
    else
      {
      found = BIL_TRUE;
      low   = mid;
      high  = mid;
      }
    }

  if (index) *index = low;
  return found;
  }


static BILboolean bilMapInsert(BILmap *map, BILint index)
  {
  bilCheckParameter(map   != NULL,        return BIL_FALSE);
  bilCheckParameter(index >= 0,           return BIL_FALSE);
  bilCheckParameter(index <= map->length, return BIL_FALSE);
  
  /* check capacity */
  if (map->capacity < (map->length+1))
    {
    if (!bilMapSetCapacity(map, map->length+1, BIL_FALSE))
      return BIL_FALSE;
    }

  /* make room for new item */
  if (index < map->length)
    memmove(&(map->items[index+1]),
            &(map->items[index  ]),
            (map->length - index) * sizeof(BILmapItem));

  /* set new list length */
  map->length++;

  return BIL_TRUE;
  }


static BILboolean bilMapRemove(BILmap *map, BILint index)
  {
  bilCheckParameter(map   != NULL,        return BIL_FALSE);
  bilCheckParameter(index >= 0,           return BIL_FALSE);
  bilCheckParameter(index <  map->length, return BIL_FALSE);
  
  /* remove old item */
  if (index < map->length)
    memmove(&(map->items[index  ]),
            &(map->items[index+1]),
            (map->length - index - 1) * sizeof(BILmapItem));

  /* set new list length */
  map->length--;

  return BIL_TRUE;
  }


static BILboolean bilMapItemInit(BILmapItem    *item,
                                 const BILchar *key,
                                 const BILvoid *value,
                                 BILsizei       size)
  {
  BILchar *newKey;
  BILvoid *newValue;

  bilCheckParameter(item  != NULL, return BIL_FALSE);
  bilCheckParameter(key   != NULL, return BIL_FALSE);
  bilCheckParameter(value != NULL, return BIL_FALSE);

  newKey = bilMemdup(key, strlen(key)+1);
  if (!newKey)
    return BIL_FALSE;

  newValue = bilMemdup(value, size);
  if (!newValue)
    {
    bilFree(newKey);
    return BIL_FALSE;
    }

  item->key   = newKey;
  item->value = newValue;
  item->size  = size;

  return BIL_TRUE;
  }


static BILboolean bilMapItemDone(BILmapItem *item)
  {
  bilCheckParameter(item != NULL, return BIL_FALSE);

  bilFree(item->key);
  bilFree(item->value);

  return BIL_TRUE;
  }


BILboolean bilMapInit(BILmap *map)
  {
  bilCheckParameter(map != NULL, return BIL_FALSE);

  map->length   = 0;
  map->capacity = 0;
  map->items    = NULL;

  return BIL_TRUE;
  }


BILboolean bilMapDone(BILmap *map)
  {
  BILboolean result;
  BILint index;  

  bilCheckParameter(map           != NULL, return BIL_FALSE);
  bilCheckParameter(map->length   >= 0,    return BIL_FALSE);
  bilCheckParameter(map->capacity >= 0,    return BIL_FALSE);

  result = BIL_TRUE;

  for (index = 0; index < map->length; index++)
    result &= bilMapItemDone(&(map->items[index]));

  bilFree(map->items);

  return result;
  }


BILboolean bilMapSetValue(BILmap        *map,
                          const BILchar *key,
                          const BILvoid *value,
                          BILsizei       size)
  {
  BILboolean found;
  BILint     index;

  bilCheckParameter(map   != NULL, return BIL_FALSE);
  bilCheckParameter(key   != NULL, return BIL_FALSE);
  bilCheckParameter(value != NULL, return BIL_FALSE);
  bilCheckParameter(size  >= 0,    return BIL_FALSE);

  found = bilMapFind(map, key, &index);
  if (found)
    {
    if (!bilMapItemDone(&(map->items[index])))
      return BIL_FALSE;
    }
  else
    {
    if (!bilMapInsert(map, index))
      return BIL_FALSE;
    }

  return bilMapItemInit(&(map->items[index]), key, value, size);
  }


const BILvoid *bilMapGetValue(const BILmap  *map,
                              const BILchar *key)
  {
  BILint     index;
  BILboolean found;

  bilCheckParameter(map != NULL, return NULL);
  bilCheckParameter(key != NULL, return NULL);

  found = bilMapFind(map, key, &index);
  if (found)
    return map->items[index].value;

  return NULL;
  }


BILsizei bilMapGetValueSize(const BILmap *map, const BILchar *key)
  {
  BILint     index;
  BILboolean found;

  bilCheckParameter(map != NULL, return 0);
  bilCheckParameter(key != NULL, return 0);

  found = bilMapFind(map, key, &index);
  if (found)
    return map->items[index].size;

  return 0;
  }


BILboolean bilMapRemoveValue(BILmap        *map,
                             const BILchar *key)
  {
  BILboolean found;
  BILint     index;
  
  bilCheckParameter(map != NULL, return BIL_FALSE);
  bilCheckParameter(key != NULL, return BIL_FALSE);

  found = bilMapFind(map, key, &index);
  if (!found)
    return BIL_FALSE;

  bilMapItemDone(&(map->items[index]));
  bilMapRemove(map, index);
  
  return BIL_TRUE;
  }


BILboolean bilMapEnumarateKeys(BILmap  *map,
                               BILchar *string,
                               BILsizei stringSize)
  {
  BILint     charPos;
  BILint     index;
  BILboolean result;

  bilCheckParameter(map        != NULL, return BIL_FALSE);
  bilCheckParameter(string     != NULL, return BIL_FALSE);
  bilCheckParameter(stringSize >= 0,    return BIL_FALSE);

  result  = BIL_TRUE;
  charPos = 0;

  for (index = 0; index < map->length; index++)
    {
    BILchar  *itemKey = map->items[index].key;
    BILsizei  itemLen = strlen(itemKey);

    if (charPos+itemLen+3 < stringSize)
      {
      /* insert space before */
      string[charPos++] = ' ';

      /* insert key */
      memcpy(&(string[charPos]), itemKey, itemLen);
      charPos += itemLen;

      /* insert space after */
      string[charPos++] = ' ';
      }
    else
      result = BIL_FALSE;
    }

  /* this will always be true, unless stringSize==0 */
  if (charPos < stringSize)
    string[charPos] = '\0';

  return result;
  }

