#include "bilInternal.h"

typedef struct
  {
  BILtype  type;
  BILubyte values;
  } BILparasiteHeader;


static BILboolean bilSetImageParasite(BILimage _image, const BILchar *key, const BILvoid *values, BILsizei size, BILtype type)
  {
  BILimageStruct    *image = _image;
  BILparasiteHeader *parasite;
  BILsizei           parasiteSize;
  BILboolean         result;

  bilCheckParameter(image  != NULL, return BIL_FALSE);
  bilCheckParameter(key    != NULL, return BIL_FALSE);
  bilCheckParameter(values != NULL, return BIL_FALSE);
  bilCheckParameter(size   >= 0,    return BIL_FALSE);

  /* allocate temporary structure */
  parasiteSize = sizeof(BILparasiteHeader) + size;
  parasite     = bilMalloc(parasiteSize);
  if (!parasite)
    return BIL_FALSE;

  /* fill in temporary structure */
  parasite->type = type;
  memcpy(&parasite->values, values, size);

  /* set parasite */
  result = bilMapSetValue(&image->parasites, key, parasite, parasiteSize);

  /* free temporary structure */
  bilFree(parasite);

  /* we are done! */
  return result;
  }


BILboolean bilSetImageParasiteBoolean(BILimage image, const BILchar *key, const BILboolean *values, BILsizei count)
  {
  return bilSetImageParasite(image, key, values, sizeof(BILboolean)*count, BIL_BOOLEAN);
  }


BILboolean bilSetImageParasiteInt(BILimage image, const BILchar *key, const BILint *values, BILsizei count)
  {
  return bilSetImageParasite(image, key, values, sizeof(BILint)*count, BIL_INT);
  }


BILboolean bilSetImageParasiteDouble(BILimage image, const BILchar *key, const BILdouble  *values, BILsizei count)
  {
  return bilSetImageParasite(image, key, values, sizeof(BILdouble)*count, BIL_DOUBLE);
  }


BILboolean bilSetImageParasiteCharPtr(BILimage image, const BILchar *key, const BILchar *value)
  {
  return bilSetImageParasite(image, key, value, sizeof(BILchar)*(strlen(value)+1), BIL_CHAR_PTR);
  }


BILboolean bilSetImageParasiteVoidPtr(BILimage image, const BILchar *key, const BILvoid *value, BILsizei size)
  {
  return bilSetImageParasite(image, key, value, size, BIL_VOID_PTR);
  }


static const BILvoid *bilGetImageParasite(BILimage _image, const BILchar *key, BILtype type)
  {
  BILimageStruct          *image = _image;
  const BILparasiteHeader *parasite;

  bilCheckParameter(image != NULL, return NULL);
  bilCheckParameter(key   != NULL, return NULL);

  /* get parasite */
  parasite = bilMapGetValue(&image->parasites, key);
  if (!parasite)
    return NULL;

  /* do some type checking */
  if (type != BIL_VOID_PTR)
    if (type != parasite->type)
      {
      bilSetError(BIL_INVALID_OPERATION);
      return NULL;
      }

  /* return values */
  return &parasite->values;
  }


const BILboolean *bilGetImageParasiteBoolean(BILimage image, const BILchar *key)
  {
  return bilGetImageParasite(image, key, BIL_BOOLEAN);
  }


const BILint *bilGetImageParasiteInt(BILimage image, const BILchar *key)
  {
  return bilGetImageParasite(image, key, BIL_INT);
  }


const BILdouble *bilGetImageParasiteDouble(BILimage image, const BILchar *key)
  {
  return bilGetImageParasite(image, key, BIL_DOUBLE);
  }


const BILchar *bilGetImageParasiteCharPtr(BILimage image, const BILchar *key)
  {
  return bilGetImageParasite(image, key, BIL_CHAR_PTR);
  }


const BILvoid *bilGetImageParasiteVoidPtr(BILimage image, const BILchar *key)
  {
  return bilGetImageParasite(image, key, BIL_VOID_PTR);
  }


BILsizei bilGetImageParasiteCount(BILimage _image, const BILchar *key)
  {
  BILimageStruct          *image = _image;
  const BILparasiteHeader *parasite;
  BILsizei                 parasiteSize;

  bilCheckParameter(image != NULL, return 0);
  bilCheckParameter(key   != NULL, return 0);

  parasiteSize = bilMapGetValueSize(&image->parasites, key) - sizeof(BILparasiteHeader);
  parasite     = bilMapGetValue    (&image->parasites, key);
  if (!parasite)
    return 0;

  switch (parasite->type)
    {
    case BIL_BOOLEAN:  return parasiteSize / sizeof(BILboolean);
    case BIL_INT:      return parasiteSize / sizeof(BILint);
    case BIL_DOUBLE:   return parasiteSize / sizeof(BILdouble);
    case BIL_CHAR_PTR: return 1;
    case BIL_VOID_PTR: return 1;
    default:           bilAssert(0, bilSetError(BIL_INVALID_ENUM));
    }

  return 0;
  }


BILsizei bilGetImageParasiteSize(BILimage _image, const BILchar *key)
  {
  BILimageStruct *image = _image;
  BILsizei        size;

  bilCheckParameter(image != NULL, return 0);
  bilCheckParameter(key   != NULL, return 0);

  size = bilMapGetValueSize(&image->parasites, key);
  if (size > 0)
    size -= sizeof(BILparasiteHeader);

  return size;
  }


BILtype bilGetImageParasiteType(BILimage _image, const BILchar *key)
  {
  BILimageStruct          *image = _image;
  const BILparasiteHeader *parasite;

  bilCheckParameter(image != NULL, return 0);
  bilCheckParameter(key   != NULL, return 0);

  parasite = bilMapGetValue(&image->parasites, key);
  if (!parasite)
    return BIL_VOID_PTR;

  return parasite->type;
  }


BILboolean bilEnumerateImageParasites(BILimage _image, BILchar *string, BILsizei stringSize)
  {
  BILimageStruct *image = _image;

  bilCheckParameter(image      != NULL, return BIL_FALSE);
  bilCheckParameter(string     != NULL, return BIL_FALSE);
  bilCheckParameter(stringSize >= 0,    return BIL_FALSE);

  return bilMapEnumarateKeys(&image->parasites, string, stringSize);
  }


BILboolean bilCopyImageParasites(BILimage src, BILimage dst)
  {
  BILboolean     ok;
  BILchar        parasites[1024];
  const BILchar *parasitesString;
  BILchar        parasite[256];

  bilCheckParameter(src != NULL, return BIL_FALSE);
  bilCheckParameter(dst != NULL, return BIL_FALSE);

  ok = bilEnumerateImageParasites(src, parasites, sizeof(parasites));
  if (!ok)
     return BIL_FALSE;

  parasitesString = parasites;
  while ((parasitesString = bilNextKey(parasitesString, parasite, sizeof(parasite))) != NULL)
    {
    switch (bilGetImageParasiteType(src, parasite))
      {
      case BIL_BOOLEAN:
        ok = bilSetImageParasiteBoolean(dst, parasite,
                                        bilGetImageParasiteBoolean(src, parasite),
                                        bilGetImageParasiteCount  (src, parasite));
        break;

      case BIL_INT:
        ok = bilSetImageParasiteInt(dst, parasite,
                                    bilGetImageParasiteInt  (src, parasite),
                                    bilGetImageParasiteCount(src, parasite));
        break;

      case BIL_DOUBLE:
        ok = bilSetImageParasiteDouble(dst, parasite,
                                       bilGetImageParasiteDouble(src, parasite),
                                       bilGetImageParasiteCount (src, parasite));
        break;

      case BIL_CHAR_PTR:
        ok = bilSetImageParasiteCharPtr(dst, parasite,
                                        bilGetImageParasiteCharPtr(src, parasite));
        break;

      case BIL_VOID_PTR:
        ok = bilSetImageParasiteVoidPtr(dst, parasite,
                                        bilGetImageParasiteVoidPtr(src, parasite),
                                        bilGetImageParasiteSize   (src, parasite));
        break;

      default:
        bilAssert(BIL_FALSE, ok = BIL_FALSE);
      }

    /* error checking */
    if (!ok)
      return BIL_FALSE;
    }

  return BIL_TRUE;
  }
