#include "bilJpegInternal.h"


typedef struct
  {
  BILdecodeContextStruct parent;
  FILE                  *tempFile;
  BILsizei               bytesDecoded;
  } BILJPEGdecodeContextStruct;


static void jpegWriteMarkersToImage(BILimage image, j_decompress_ptr cinfo)
  {
  jpeg_saved_marker_ptr marker;
  BILchar               key[32];
  BILchar               comment[0xFFFF];
  BILsizei              commentSize;

  for (marker = cinfo->marker_list; marker != NULL; marker = marker->next)
    {
    if (marker->marker == JPEG_COM)
      {
      /* store comments in "comment" parasite */
      sprintf(key, "comment");

      /* there could be multiple comment markers! */
      commentSize = bilGetImageParasiteSize(image, key);
      if (commentSize > 0)
        {
        /* get existing comment */
        memcpy(comment, bilGetImageParasiteCharPtr(image, key), commentSize);

        /* append '\n' */
        comment[commentSize++] = '\n';
        }

      /* append new comment */
      memcpy(comment+commentSize, marker->data, marker->data_length);
      commentSize += marker->data_length;

      /* append '\0' if needed */
      if (comment[commentSize-1] != '\0')
        comment[commentSize++] = '\0';

      /* set it... */
      bilSetImageParasiteCharPtr(image, key, comment);
      }
    else
      {
      /* APP0 to APP15 markers are simply saved as-is... */
      sprintf(key, "jpeg_app%d", marker->marker - JPEG_APP0);
      bilSetImageParasiteVoidPtr(image, key, marker->data, marker->data_length);
      }
    }
  }


static BILcontext bilJpegNewDecodeContext()
  {
  BILJPEGdecodeContextStruct *context;

  context = bilMalloc(sizeof(BILJPEGdecodeContextStruct));
  if (!context)
    return NULL;

  return (BILcontext)context;
  }


static BILboolean bilJpegDeleteDecodeContext(BILcontext _context)
  {
  BILJPEGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  bilFree(context);

  return BIL_TRUE;
  }


static BILboolean bilJpegBeginDecode(BILcontext _context)
  {
  BILJPEGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  /* begin decoding */  
  context->bytesDecoded = 0;
  context->tempFile     = tmpfile();
  if (!context->tempFile)
     {
     bilSetError(BIL_FILE_WRITE_ERROR);
     return BIL_FALSE;
     }

  return BIL_TRUE;
  }


static BILboolean bilJpegEndDecode(BILcontext _context)
  {
  BILJPEGdecodeContextStruct    *context  = _context;
  volatile BILboolean            result   = BIL_FALSE;
  JSAMPROW                       scanline = NULL;
  BILJPEGerrorMgrStruct          errorMgr;
  struct jpeg_decompress_struct  cinfo;
  JDIMENSION                     num_scanlines;
  BILformat                      format;
  BILtype                        type;
  BILsizei                       bpp;
  BILint                         marker;
  BILboolean                     progressive;

  bilCheckParameter(context               != NULL, return BIL_FALSE);
  bilCheckParameter(context->parent.image != NULL, return BIL_FALSE);
  bilCheckParameter(context->tempFile     != NULL, return BIL_FALSE);

  /* initialise error manager */
  if (!bilJpegInitErrorMgr(&errorMgr))
    return BIL_FALSE;

  /* catch all errors */
  if (setjmp(errorMgr.setjmp_buffer))
    {
    bilSetError(BIL_INVALID_OPERATION);
    goto error_end;
    }

  /* initialize the JPEG decompression object */
  cinfo.err = &errorMgr.parent;
  jpeg_create_decompress(&cinfo);

  /* save all markers */
  jpeg_save_markers(&cinfo, JPEG_COM, 0xFFFF);
  for (marker = 0; marker < 16; marker++)
    jpeg_save_markers(&cinfo, JPEG_APP0+marker, 0xFFFF);

  /* specify data source for decompression */
  rewind(context->tempFile);
  jpeg_stdio_src(&cinfo, context->tempFile);

  /* read file header, set default decompression parameters */
  jpeg_read_header(&cinfo, TRUE);

  /* save markers in image */
  jpegWriteMarkersToImage(context->parent.image, &cinfo);

  /* save jpeg parameters (the ones we know about) */
  progressive = (cinfo.progressive_mode) ? BIL_TRUE : BIL_FALSE;
  bilSetImageParasiteBoolean(context->parent.image, "jpeg_progressive", &progressive, 1);

  /* start decompressor */
  jpeg_start_decompress(&cinfo);

  /* get image format */
  switch (cinfo.output_components)
    {
    case 1: /* grayscale image */
      bpp      = 1;
      format   = BIL_LUMINANCE;
      type     = BIL_UNSIGNED_BYTE;      
      break;

    case 2: /* ga data (not yet possible) */
      bpp    = 2;
      format = BIL_LUMINANCE_ALPHA;
      type   = BIL_UNSIGNED_BYTE;
      break;

    case 3: /* rgb data */
      bpp    = 3;
      format = BIL_RGB;
      type   = BIL_UNSIGNED_BYTE;
      break;

    case 4: /* rgba data (not yet possible) */
      bpp    = 4;
      format = BIL_RGBA;
      type   = BIL_UNSIGNED_BYTE;
      break;

    default:
      /* should never happen! */
      bilSetError(BIL_INVALID_OPERATION);
      goto error_end;
    }

  /* allocate scanline */
  scanline = bilMalloc(cinfo.output_width * bpp);
  if (!scanline)
    goto error_end;

  /* resize image */
  if (!bilResizeImage(context->parent.image, cinfo.output_width, cinfo.output_height))
    goto error_end;

  /* process data */
  while (cinfo.output_scanline < cinfo.output_height)
    {
    BILint y = cinfo.output_scanline;

    num_scanlines = jpeg_read_scanlines(&cinfo, &scanline, 1);
    bilDrawPixels(context->parent.image, 0, y, cinfo.output_width, num_scanlines, format, type, scanline);
    }

  /* finish decompression */
  jpeg_finish_decompress(&cinfo);

  /* we were successful */
  result = BIL_TRUE;

error_end:

  /* release memory */
  bilFree(scanline);
  jpeg_destroy_decompress(&cinfo);

  /* close files */
  if (fclose(context->tempFile) < 0)
    {
    bilSetError(BIL_FILE_WRITE_ERROR);
    return BIL_FALSE;
    }

  return result;
  }


static BILsizei bilJpegDecodeBlock(BILcontext _context, const BILvoid *block, BILsizei size)
  {
  BILJPEGdecodeContextStruct *context = _context;
  BILsizei                    numWritten;

  bilCheckParameter(context != NULL, return 0);
  bilCheckParameter(block   != NULL, return 0);
  bilCheckParameter(size    >= 0,    return 0);

  /* write block to temporary file */
  numWritten = fwrite(block, 1, size, context->tempFile);
  if (numWritten != size)
    {
    bilSetError(BIL_FILE_WRITE_ERROR);
    return 0;
    }

  context->bytesDecoded += numWritten;

  return numWritten;
  }


static BILboolean bilJpegValidBlock(BILcontext _context, const BILvoid *block, BILsizei size)
  {
  BILJPEGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);
  bilCheckParameter(block   != NULL, return BIL_FALSE);
  bilCheckParameter(size    >= 0,    return BIL_FALSE);

  /* a zero sized block is NOT valid */
  if (size <= 0)
    return BIL_FALSE;

  return BIL_TRUE;
  }


BILboolean bilJpegRegisterDecoder()
  {
  static BILparameterStruct parameters[] =
        {
        { NULL, 0, 0 }
        };

  static BILdecoderStruct decoder =
        {
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BILJPEG_VERSION,
        /* parameters          */ parameters,
        /* newDecodeContext    */ bilJpegNewDecodeContext,
        /* deleteDecodeContext */ bilJpegDeleteDecodeContext,
        /* beginDecode         */ bilJpegBeginDecode,
        /* endDecode           */ bilJpegEndDecode,
        /* decodeBlock         */ bilJpegDecodeBlock,
        /* validBlock          */ bilJpegValidBlock
        };

  return bilRegisterDecoder(BILJPEG_NAME, &decoder);
  }


BILboolean bilJpegUnregisterDecoder()
  {
  return bilUnregisterDecoder(BILJPEG_NAME);
  }

