#include "bilJpegInternal.h"


typedef struct
  {
  BILencodeContextStruct  parent;
  FILE                   *tempFile;
  BILsizei                bytesToEncode;
  BILsizei                bytesEncoded;
  } BILJPEGencodeContextStruct;


static void jpegWriteMarkersToJpeg(BILimage image, j_compress_ptr cinfo)
  {
  BILchar        key[32];
  const BILchar *parasitePtr;
  BILsizei       parasiteSize;
  BILint         marker;

  /* write comment marker */
  sprintf(key, "comment");
  parasitePtr  = bilGetImageParasiteCharPtr(image, key);
  parasiteSize = bilGetImageParasiteSize(image, key);
  if ((parasitePtr != NULL) && (parasiteSize > 0))
    jpeg_write_marker(cinfo, JPEG_COM, parasitePtr, parasiteSize);

  /* write application markers */
  for (marker = 0; marker < 16; marker++)
    {
    /* dont write JFIF marker (library does it for us) */
    if (marker == 0) continue;

    /* dont write Adobe marker (library does it for us) */
    if (marker == 14) continue;

    /* marker is stored in jpeg_appN property */
    sprintf(key, "jpeg_app%d", marker);
    parasitePtr  = bilGetImageParasiteVoidPtr(image, key);
    parasiteSize = bilGetImageParasiteSize(image, key);
    if ((parasitePtr != NULL) && (parasiteSize > 0))
      jpeg_write_marker(cinfo, JPEG_APP0+marker, parasitePtr, parasiteSize);
    }
  }


static BILcontext bilJpegNewEncodeContext()
  {
  BILJPEGencodeContextStruct *context;

  context = bilMalloc(sizeof(BILJPEGencodeContextStruct));
  if (!context)
    return NULL;

  return (BILcontext)context;
  }


static BILboolean bilJpegDeleteEncodeContext(BILcontext _context)
  {
  BILJPEGencodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  bilFree(context);

  return BIL_TRUE;
  }


static BILboolean bilJpegBeginEncode(BILcontext _context)
  {
  BILJPEGencodeContextStruct    *context  = _context;
  volatile BILboolean            result   = BIL_FALSE;
  JSAMPROW                       scanline = NULL;
  BILJPEGerrorMgrStruct          errorMgr;
  struct jpeg_compress_struct    cinfo;
  const BILdouble               *qualityPtr;
  const BILdouble               *smoothingPtr;
  const BILboolean              *optimizePtr;
  const BILboolean              *progressivePtr;
  const BILboolean              *baselinePtr;

  bilCheckParameter(context               != NULL, return BIL_FALSE);
  bilCheckParameter(context->parent.image != NULL, return BIL_FALSE);

  /* initialise error manager */  
  if (!bilJpegInitErrorMgr(&errorMgr))
    return BIL_FALSE;

  /* create temporary output file */
  context->tempFile = tmpfile();
  if (!context->tempFile)
    {
    bilSetError(BIL_FILE_WRITE_ERROR);
    goto error_end;
    }

  /* catch all errors */
  if (setjmp(errorMgr.setjmp_buffer))
    {
    bilSetError(BIL_INVALID_OPERATION);
    goto error_end;
    }

  /* initialize the JPEG compression object */
  cinfo.err = &errorMgr.parent;
  jpeg_create_compress(&cinfo);

  /* specify data destination */
  jpeg_stdio_dest(&cinfo, context->tempFile);

  /* set image details */
  cinfo.image_width      = bilGetImageWidth (context->parent.image);
  cinfo.image_height     = bilGetImageHeight(context->parent.image);
  cinfo.input_components = 3;
  cinfo.in_color_space   = JCS_RGB;

  /* get compression parameters */
  qualityPtr     = bilGetImageParasiteDouble (context->parent.image, "jpeg_quality");
  smoothingPtr   = bilGetImageParasiteDouble (context->parent.image, "jpeg_smoothing");
  optimizePtr    = bilGetImageParasiteBoolean(context->parent.image, "jpeg_optimize");
  progressivePtr = bilGetImageParasiteBoolean(context->parent.image, "jpeg_progressive");
  baselinePtr    = bilGetImageParasiteBoolean(context->parent.image, "jpeg_baseline");

  /* set compression parameters */
  jpeg_set_defaults(&cinfo);
  if (qualityPtr)     jpeg_set_quality(&cinfo, (int)(*qualityPtr   * 100), (boolean)((baselinePtr) ? *baselinePtr : BIL_TRUE));
  if (smoothingPtr)   cinfo.smoothing_factor = (int)(*smoothingPtr * 100);
  if (optimizePtr)    cinfo.optimize_coding  = *optimizePtr;
  if (progressivePtr) if (*progressivePtr) jpeg_simple_progression(&cinfo);

  /* allocate scanline conversion buffer */
  scanline = bilMalloc(cinfo.image_width * cinfo.input_components);
  if (!scanline)
    goto error_end;

  /* start compression */
  jpeg_start_compress(&cinfo, TRUE);

  /* write image markers */
  jpegWriteMarkersToJpeg(context->parent.image, &cinfo);

  /* process data */
  while (cinfo.next_scanline < cinfo.image_height)
    {
    BILint y = cinfo.next_scanline;

    bilReadPixels(context->parent.image, 0, y, cinfo.image_width, 1, BIL_RGB, BIL_UNSIGNED_BYTE, scanline);
    jpeg_write_scanlines(&cinfo, &scanline, 1);
    }

  /* finish decompression */
  jpeg_finish_compress(&cinfo);

  /* set some more details */
  context->bytesToEncode = ftell(context->tempFile);
  context->bytesEncoded  = 0;

  /* rewind file */
  rewind(context->tempFile);

  /* we were successful */
  result = BIL_TRUE;  

error_end:

  /* release memory */
  jpeg_destroy_compress(&cinfo);
  bilFree(scanline);

  return result;
  }


static BILboolean bilJpegEndEncode(BILcontext _context)
  {
  BILJPEGencodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  /* close file */
  if (fclose(context->tempFile) < 0)
    {
    bilSetError(BIL_FILE_READ_ERROR);
    return BIL_FALSE;
    }

  /* test to make sure entire image was encoded */
  if (context->bytesEncoded != context->bytesToEncode)
    {
    bilSetError(BIL_INVALID_OPERATION);
    return BIL_FALSE;
    }

  return BIL_TRUE;
  }


static BILsizei bilJpegEncodeBlock(BILcontext _context, BILvoid *block, BILsizei size)
  {
  BILJPEGencodeContextStruct *context = _context;
  BILsizei                    numRead;

  bilCheckParameter(context != NULL, return 0);
  bilCheckParameter(block   != NULL, return 0);
  bilCheckParameter(size    >= 0,    return 0);

  /* read next block */
  numRead = fread(block, 1, size, context->tempFile);

  /* check for read error */
  if (ferror(context->tempFile))
    {
    bilSetError(BIL_FILE_READ_ERROR);
    return 0;
    }

  /* increment encoding counter */
  context->bytesEncoded += numRead;

  return numRead;
  }


BILboolean bilJpegRegisterEncoder()
  {
  static BILparameterStruct parameters[] =
        {
        { "comment",          BIL_CHAR_PTR, 1 },
        { "jpeg_quality",     BIL_FLOAT,    1 },
        { "jpeg_smoothing",   BIL_FLOAT,    1 },
        { "jpeg_optimize",    BIL_BOOLEAN,  1 },
        { "jpeg_progressive", BIL_BOOLEAN,  1 },
        { "jpeg_baseline",    BIL_BOOLEAN,  1 },
        { NULL,               0,            0 }
        };

  static BILencoderStruct encoder =
        {
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BILJPEG_VERSION,
        /* parameters          */ parameters,
        /* newEncodeContext    */ bilJpegNewEncodeContext,
        /* deleteEncodeContext */ bilJpegDeleteEncodeContext,
        /* beginEncode         */ bilJpegBeginEncode,
        /* endEncode           */ bilJpegEndEncode,
        /* encodeBlock         */ bilJpegEncodeBlock
        };

  return bilRegisterEncoder(BILJPEG_NAME, &encoder);
  }


BILboolean bilJpegUnregisterEncoder()
  {
  return bilUnregisterEncoder(BILJPEG_NAME);
  }

