#include "bilPngInternal.h"

typedef struct
  {
  BILdecodeContextStruct parent;
  png_structp            png_ptr;
  png_infop              info_ptr;
  BILsizei               width;
  BILsizei               height;
  BILformat              format;
  BILtype                type;
  BILboolean             finished;
  } BILPNGdecodeContextStruct;


static void pngWriteCommentsToImage(BILimage image, png_structp png_ptr, png_infop info_ptr)
  {
  BILchar   comment[0xFFFF];
  BILsizei  commentSize;
  BILsizei  numComments;  
  png_text *textPtr;
  BILint    index;

  numComments = png_get_text(png_ptr, info_ptr, &textPtr, NULL);
  if (numComments <= 0)
    return;

  commentSize = bilGetImageParasiteSize(image, "comment");
  if (commentSize > 0)
    memcpy(comment, bilGetImageParasiteCharPtr(image, "comment"), commentSize); 

  for (index = 0; index < numComments; index++)
    {
    if (textPtr[index].text_length >= 0)
      {
      /* append '\n' (if we aleady have a comment) */
      if (commentSize > 0)
        comment[commentSize++] = '\n';

      /* append key to comment */
      memcpy(comment+commentSize, textPtr[index].key, strlen(textPtr[index].key));
      commentSize += strlen(textPtr[index].key);

      /* colon separates text from key */
      comment[commentSize++] = ':';
      comment[commentSize++] = ' ';

      /* append text to comment */
      memcpy(comment+commentSize, textPtr[index].text, textPtr[index].text_length);
      commentSize += textPtr[index].text_length;
      }
    }

  /* did we have any comments? */
  if (commentSize <= 0)
    return;

  /* append '\0' if needed */
  if (comment[commentSize-1] != '\0')
    comment[commentSize++] = '\0';

  /* set comment parasite */
  bilSetImageParasiteCharPtr(image, "comment", comment);
  }


static void pngInfoCallback(png_structp png_ptr, png_infop info_ptr)
  {
  BILPNGdecodeContextStruct *context = png_get_progressive_ptr(png_ptr);
  const BILdouble           *gammaPtr;
  BILchar                   *gammaStr;
  BILdouble                  screenGamma;
  BILdouble                  imageGamma;
  BILint                     intent;

  bilCheckParameter(context               != NULL, return);
  bilCheckParameter(context->parent.image != NULL, return);

  /* get image information */
  context->width  = png_get_image_width(png_ptr, info_ptr);
  context->height = png_get_image_height(png_ptr, info_ptr);  
  context->format = BIL_RGB;
  context->type   = BIL_UNSIGNED_BYTE;

  /* get image format */
  switch (png_get_color_type(png_ptr, info_ptr))
    {
    case PNG_COLOR_TYPE_GRAY:       context->format = BIL_LUMINANCE;       break;
    case PNG_COLOR_TYPE_PALETTE:    context->format = BIL_RGB;             break;
    case PNG_COLOR_TYPE_RGB:        context->format = BIL_RGB;             break;
    case PNG_COLOR_TYPE_RGB_ALPHA:  context->format = BIL_RGBA;            break;
    case PNG_COLOR_TYPE_GRAY_ALPHA: context->format = BIL_LUMINANCE_ALPHA; break;
    default:                        bilAssert(0, longjmp(png_jmpbuf((context->png_ptr)), 1));
    }

  /* resize image */
  if (!bilResizeImage(context->parent.image, context->width, context->height))
    longjmp(png_jmpbuf((context->png_ptr)), 1);

  /* clear image */
  if (!bilClearImage(context->parent.image))
    longjmp(png_jmpbuf((context->png_ptr)), 1);

  /* get screen gamma */
  if ((gammaPtr = bilGetImageParasiteDouble(context->parent.image, "screen_gamma")) != NULL)
    screenGamma = *gammaPtr;
  else if ((gammaStr = getenv("SCREEN_GAMMA")) != NULL)
    screenGamma = strtod(gammaStr, NULL);
  else
    screenGamma = 2.2;

  /* get image gamma */
  if (png_get_sRGB(png_ptr, info_ptr, &intent))
    imageGamma = 0.45455;
  else
    {
    if (!png_get_gAMA(png_ptr, info_ptr, &imageGamma))
      imageGamma = 0.45455;
    }

  /* save gamma values in image */
  bilSetImageParasiteDouble(context->parent.image, "screen_gamma", &screenGamma, 1);
  bilSetImageParasiteDouble(context->parent.image, "image_gamma",  &imageGamma,  1);

  /* please do the gamma correction for us */
  png_set_gamma(png_ptr, screenGamma, imageGamma);

  /* strip 16 bit/color files down to 8 bits/color */
  png_set_strip_16(png_ptr);

  /* expand data to useful type */
  png_set_expand(png_ptr);

  /* handle interlaced images automatically */
  png_set_interlace_handling(png_ptr);

  /* here we go... */
  png_start_read_image(png_ptr);
  }


static void pngRowCallback(png_structp png_ptr, png_bytep new_row, png_uint_32 row_num, int pass)
  {
  BILPNGdecodeContextStruct *context = png_get_progressive_ptr(png_ptr);

  bilCheckParameter(context               != NULL, return);
  bilCheckParameter(context->parent.image != NULL, return);

  if (new_row)
    bilDrawPixels(context->parent.image, 0, row_num, context->width, 1, context->format, context->type, new_row);
  }


static void pngEndCallback(png_structp png_ptr, png_infop info_ptr)
  {
  BILPNGdecodeContextStruct *context = png_get_progressive_ptr(png_ptr);

  bilCheckParameter(context               != NULL, return);
  bilCheckParameter(context->parent.image != NULL, return);

  /* we have finished reading the image */
  context->finished = BIL_TRUE;

  /* now, grab the comments */
  pngWriteCommentsToImage(context->parent.image, png_ptr, info_ptr);
  }


static BILcontext bilPngNewDecodeContext()
  {
  BILPNGdecodeContextStruct *context;

  context = bilMalloc(sizeof(BILPNGdecodeContextStruct));
  if (!context)
    return NULL;

  return (BILcontext)context;
  }


static BILboolean bilPngDeleteDecodeContext(BILcontext _context)
  {
  BILPNGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  bilFree(context);

  return BIL_TRUE;
  }


static BILboolean bilPngBeginDecode(BILcontext _context)
  {
  BILPNGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  context->finished = BIL_FALSE;

  context->png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!context->png_ptr)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    return BIL_FALSE;
    }

  context->info_ptr = png_create_info_struct(context->png_ptr);
  if (!context->info_ptr)
    {
    png_destroy_read_struct(&context->png_ptr, &context->info_ptr, (png_infopp)NULL);
    bilSetError(BIL_OUT_OF_MEMORY);
    return BIL_FALSE;
    }

  if (setjmp(png_jmpbuf((context->png_ptr))))
    {
    png_destroy_read_struct(&context->png_ptr, &context->info_ptr, (png_infopp)NULL);
    bilSetError(BIL_INVALID_OPERATION);
    return BIL_FALSE;
    }

  png_set_progressive_read_fn(context->png_ptr, context,
                              pngInfoCallback, pngRowCallback, pngEndCallback);

  return BIL_TRUE;
  }


static BILboolean bilPngEndDecode(BILcontext _context)
  {
  BILPNGdecodeContextStruct *context = _context;

  bilCheckParameter(context               != NULL, return BIL_FALSE);
  bilCheckParameter(context->parent.image != NULL, return BIL_FALSE);

  /* destroy the png structures */
  png_destroy_read_struct(&context->png_ptr, &context->info_ptr, (png_infopp)NULL);

  /* did we finish decoding the image? */
  if (!context->finished)
    {
    bilSetError(BIL_INVALID_OPERATION);
    return BIL_FALSE;
    }

  return BIL_FALSE;
  }


static BILsizei bilPngDecodeBlock(BILcontext _context, const BILvoid *block, BILsizei size)
  {
  BILPNGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return 0);
  bilCheckParameter(block   != NULL, return 0);
  bilCheckParameter(size    >= 0,    return 0);

  if (setjmp(png_jmpbuf((context->png_ptr))))
    {
    bilSetError(BIL_INVALID_OPERATION);
    return 0;
    }

  png_process_data(context->png_ptr, context->info_ptr, (png_bytep)block, (png_size_t)size);
  
  return size;
  }


static BILboolean bilPngValidBlock(BILcontext _context, const BILvoid *block, BILsizei size)
  {
  BILPNGdecodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);
  bilCheckParameter(block   != NULL, return BIL_FALSE);
  bilCheckParameter(size    >= 0,    return BIL_FALSE);

  /* a zero sized block is NOT valid */
  if (size <= 0)
    return BIL_FALSE;

  /* if the image is already decoded, then the block is not valid */
  if (context->finished)
    return BIL_FALSE;

  return BIL_TRUE;
  }


BILboolean bilPngRegisterDecoder()
  {
  static BILparameterStruct parameters[] =
        {
        { "screen_gamma", BIL_DOUBLE, 1 },
        { NULL,           0,          0 }
        };

  static BILdecoderStruct decoder = 
        {
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BILPNG_VERSION,
        /* parameters          */ parameters,
        /* newDecodeContext    */ bilPngNewDecodeContext,
        /* deleteDecodeContext */ bilPngDeleteDecodeContext,
        /* beginDecode         */ bilPngBeginDecode,
        /* endDecode           */ bilPngEndDecode,
        /* decodeBlock         */ bilPngDecodeBlock,
        /* validBlock          */ bilPngValidBlock
        };

  return bilRegisterDecoder(BILPNG_NAME, &decoder);
  }


BILboolean bilPngUnregisterDecoder()
  {
  return bilUnregisterDecoder(BILPNG_NAME);
  }

