#include "bilPngInternal.h"


typedef struct
  {
  BILencodeContextStruct  parent;
  FILE                   *tempFile;
  BILsizei                bytesToEncode;
  BILsizei                bytesEncoded;
  } BILPNGencodeContextStruct;


static BILcontext bilPngNewEncodeContext()
  {
  BILPNGencodeContextStruct *context;

  context = bilMalloc(sizeof(BILPNGencodeContextStruct));
  if (!context)
    return NULL;

  return (BILcontext)context;
  }


static BILboolean bilPngDeleteEncodeContext(BILcontext _context)
  {
  BILPNGencodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  bilFree(context);

  return BIL_TRUE;
  }


static BILboolean bilPngBeginEncode(BILcontext _context)
  {
  BILPNGencodeContextStruct *context      = _context;
  volatile BILboolean        result       = BIL_FALSE;  
  png_structp                png_ptr      = NULL;
  png_infop                  info_ptr     = NULL;
  volatile png_bytepp        rowPointers  = NULL;
  const BILdouble           *gammaPtr     = NULL;
  const BILchar             *comment      = NULL;
  const BILboolean          *interlacePtr = NULL;
  BILsizei                   width;
  BILsizei                   height;
  BILint                     row;

  bilCheckParameter(context               != NULL, return BIL_FALSE);
  bilCheckParameter(context->parent.image != NULL, return BIL_FALSE);

  width  = bilGetImageWidth(context->parent.image);
  height = bilGetImageHeight(context->parent.image);

  /* create temporary file */
  context->tempFile = tmpfile();
  if (!context->tempFile)
    {
    bilSetError(BIL_FILE_WRITE_ERROR);
    goto error_end;
    }

  /* create png writer */
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    goto error_end;
    }

  /* create png info */
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    goto error_end;
    }

  /* png error handler */
  if (setjmp(png_jmpbuf(png_ptr)))
    {
    bilSetError(BIL_INVALID_OPERATION);
    goto error_end;
    }

  /* allocate space for rows */
  rowPointers = bilMalloc(height * sizeof(BILvoid*));
  if (!rowPointers)
    goto error_end;

  /* writepng to temporary file */
  png_init_io(png_ptr, context->tempFile);

  /* set header */
  png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB_ALPHA,
               PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

  /* set image gamma */
  if ((gammaPtr = bilGetImageParasiteDouble(context->parent.image, "image_gamma")) != NULL)
    png_set_gAMA(png_ptr, info_ptr, *gammaPtr);

  /* set image comment */
  if ((comment = bilGetImageParasiteCharPtr(context->parent.image, "comment")) != NULL)
    {
    png_text textPtr[1];

    textPtr[0].key         = (png_charp)"Comment";
    textPtr[0].text        = (png_charp)comment;
    textPtr[0].compression = (bilGetImageParasiteSize(context->parent.image, "comment") < 256) ? PNG_TEXT_COMPRESSION_NONE : PNG_TEXT_COMPRESSION_zTXt;
  #ifdef PNG_iTXt_SUPPORTED
    textPtr[0].lang        = NULL;
  #endif

    png_set_text(png_ptr, info_ptr, textPtr, 1);
    }

  /* write header */
  png_write_info(png_ptr, info_ptr);

  /* set interlacing */
  interlacePtr = bilGetImageParasiteBoolean(context->parent.image, "png_interlace");
  if (interlacePtr && *interlacePtr)
    png_set_interlace_handling(png_ptr);

  /* write image */
  for (row = 0; row < height; row++)
    rowPointers[row] = (png_bytep)bilGetImagePtr(context->parent.image, 0, row);
  png_write_image(png_ptr, rowPointers);

  /* finish writing */
  png_write_end(png_ptr, info_ptr);

  /* set some more details */
  context->bytesToEncode = ftell(context->tempFile);
  context->bytesEncoded  = 0;

  /* rewind file */
  rewind(context->tempFile);

  /* we were successful */
  result = BIL_TRUE;

error_end:

  /* release memory */
  png_destroy_write_struct(&png_ptr, &info_ptr);
  bilFree(rowPointers);

  return result;
  }


static BILboolean bilPngEndEncode(BILcontext _context)
  {
  BILPNGencodeContextStruct *context = _context;

  bilCheckParameter(context != NULL, return BIL_FALSE);

  /* close file */
  if (fclose(context->tempFile) < 0)
    {
    bilSetError(BIL_FILE_READ_ERROR);
    return BIL_FALSE;
    }

  /* test to make sure entire image was encoded */
  if (context->bytesEncoded != context->bytesToEncode)
    {
    bilSetError(BIL_INVALID_OPERATION);
    return BIL_FALSE;
    }

  return BIL_TRUE;
  }


static BILsizei bilPngEncodeBlock(BILcontext _context, BILvoid *block, BILsizei size)
  {
  BILPNGencodeContextStruct *context = _context;
  BILsizei                   numRead;

  bilCheckParameter(context != NULL, return 0);
  bilCheckParameter(block   != NULL, return 0);
  bilCheckParameter(size    >= 0,    return 0);

  /* read next block */
  numRead = fread(block, 1, size, context->tempFile);

  /* check for read error */
  if (ferror(context->tempFile))
    {
    bilSetError(BIL_FILE_READ_ERROR);
    return 0;
    }

  /* increment encoding counter */
  context->bytesEncoded += numRead;

  return numRead;
  }


BILboolean bilPngRegisterEncoder()
  {
  static BILparameterStruct parameters[] =
        {
        { "comment",       BIL_CHAR_PTR, 1 },
        { "png_interlace", BIL_BOOLEAN,  1 },
        { NULL,            0,            0 }
        };

  static BILencoderStruct encoder =
        {
        /* bilVersion          */ BIL_VERSION,
        /* version             */ BILPNG_VERSION,
        /* parameters          */ parameters,
        /* newEncodeContext    */ bilPngNewEncodeContext,
        /* deleteEncodeContext */ bilPngDeleteEncodeContext,
        /* beginEncode         */ bilPngBeginEncode,
        /* endEncode           */ bilPngEndEncode,
        /* encodeBlock         */ bilPngEncodeBlock
        };

  return bilRegisterEncoder(BILPNG_NAME, &encoder);
  }



BILboolean bilPngUnregisterEncoder()
  {
  return bilUnregisterEncoder(BILPNG_NAME);
  }

