/*
 * The pnm reading and writing code was written from scratch by Erik Nygren
 * (nygren@mit.edu) based on the specifications in the man pages and
 * does not contain any code from the netpbm or pbmplus distributions.
 *
 * Source modified by Bernhard Tschirren (bernie-t@geocities.com) to
 * slot into BIL v1.0. Source taken from "pnm.c" in the GIMP.
 */

#include "bilPpmInternal.h"

#define BUFLEN 512          /* The input buffer size for data returned
                             * from the scanner.  Note that lines
                             * aren't allowed to be over 256 characters
                             * by the spec anyways so this shouldn't
                             * be an issue. */

typedef struct _PNMScanner
  {
  FILE   *file;             /* The file being read */
  BILchar cur;              /* The current character in the input stream */
  } PNMScanner;

typedef struct _PNMInfo
  {
  BILsizei   xres, yres;    /* The size of the image */
  BILsizei   maxval;        /* For ascii image files, the max value
                             * which we need to normalize to */
  BILsizei   np;            /* Number of image planes (0 for pbm) */
  BILboolean asciibody;     /* 1 if ascii body, 0 if raw body */
  jmp_buf    jmpbuf;        /* Where to jump to on an error loading */
                            /* Routine to use to load the pnm body */
  void     (*loader)(PNMScanner *, struct _PNMInfo *, BILimage);
  } PNMInfo;

typedef struct _PNMRowInfo
  {
  FILE     *file;           /* File descriptor */
  BILsizei  xres;           /* X resolution */
  BILsizei  np;             /* Number of planes */
  } PNMRowInfo;


static void pnm_load_ascii   (PNMScanner *scan, PNMInfo *info, BILimage image);
static void pnm_load_raw     (PNMScanner *scan, PNMInfo *info, BILimage image);
static void pnm_load_rawpbm  (PNMScanner *scan, PNMInfo *info, BILimage image);

static void pnmsaverow_ascii (PNMRowInfo *ri, BILubyte *data);
static void pnmsaverow_raw   (PNMRowInfo *ri, BILubyte *data);

static PNMScanner *pnmscanner_create (FILE *file);
static void        pnmscanner_destroy(PNMScanner *s);

static void pnmscanner_getchar       (PNMScanner *s);
static void pnmscanner_eatwhitespace (PNMScanner *s);
static void pnmscanner_gettoken      (PNMScanner *s, BILchar *buf, BILsizei bufsize);
static void pnmscanner_getsmalltoken (PNMScanner *s, BILchar *buf);

#define pnmscanner_eof(s)   (feof((s)->file))
#define pnmscanner_file(s)  ((s)->file)

/* Checks for a fatal error */
#define CHECK_FOR_ERROR(predicate, jmpbuf, error)  \
        if ((predicate))                           \
        { bilSetError(error); longjmp((jmpbuf),1); }

static struct struct_pnm_types
  {
  BILchar  name;
  BILint   np;
  BILint   asciibody;
  BILint   maxval;
  void   (*loader)(PNMScanner *, struct _PNMInfo *, BILimage image);
  } pnm_types[] = {
                  { '1', 0, 1,   1, pnm_load_ascii },  /* ASCII PBM */
                  { '2', 1, 1, 255, pnm_load_ascii },  /* ASCII PGM */
                  { '3', 3, 1, 255, pnm_load_ascii },  /* ASCII PPM */
                  { '4', 0, 0,   1, pnm_load_rawpbm }, /* RAW   PBM */
                  { '5', 1, 0, 255, pnm_load_raw },    /* RAW   PGM */
                  { '6', 3, 0, 255, pnm_load_raw },    /* RAW   PPM */
                  {  0 , 0, 0,   0, NULL}
                  };


BILboolean ppm_load_image(BILPPMdecodeContextStruct *context)
  {
  BILchar     buf[BUFLEN];
  PNMInfo    *pnminfo;
  PNMScanner *scan;
  BILint      ctr;

  /* create pnm info */
  pnminfo = malloc(sizeof(PNMInfo));
  if (!pnminfo)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    return BIL_FALSE;
    }

  /* create scanner */
  scan = pnmscanner_create(context->tempFile);
  if (!scan)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    free(pnminfo);
    return BIL_FALSE;
    }

  /* set error handling */
  if (setjmp(pnminfo->jmpbuf))
    {
    pnmscanner_destroy(scan);
    free(pnminfo);
    return BIL_FALSE;
    }

  /* get magic number */
  pnmscanner_gettoken(scan, buf, BUFLEN);
  CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf, BIL_FILE_READ_ERROR);
  CHECK_FOR_ERROR((buf[0] != 'P' || buf[2]), pnminfo->jmpbuf, BIL_INVALID_OPERATION);

  /* Look up magic number to see what type of PNM this is */
  for (ctr=0; pnm_types[ctr].name; ctr++)
    {
    if (buf[1] == pnm_types[ctr].name)
      {
      pnminfo->np        = pnm_types[ctr].np;
      pnminfo->asciibody = pnm_types[ctr].asciibody;
      pnminfo->maxval    = pnm_types[ctr].maxval;
      pnminfo->loader    = pnm_types[ctr].loader;

      if (pnminfo->asciibody)
        {
        BILboolean true = BIL_TRUE;
        bilSetImageParasiteBoolean(context->parent.image, "ppm_ascii", &true, 1);
        }
      }
    }
  CHECK_FOR_ERROR((!pnminfo->loader), pnminfo->jmpbuf, BIL_INVALID_OPERATION);
  
  /* get width */
  pnmscanner_gettoken(scan, buf, BUFLEN);
  CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf, BIL_FILE_READ_ERROR);
  pnminfo->xres = isdigit(*buf)?atoi(buf):0;
  CHECK_FOR_ERROR(pnminfo->xres<=0, pnminfo->jmpbuf, BIL_INVALID_OPERATION);

  /* get height */
  pnmscanner_gettoken(scan, buf, BUFLEN);
  CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf, BIL_FILE_READ_ERROR);
  pnminfo->yres = isdigit(*buf)?atoi(buf):0;
  CHECK_FOR_ERROR(pnminfo->yres<=0, pnminfo->jmpbuf, BIL_INVALID_OPERATION);

  if (pnminfo->np != 0)		/* pbm's don't have a maxval field */
    {
    pnmscanner_gettoken(scan, buf, BUFLEN);
    CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf, BIL_FILE_READ_ERROR);

    pnminfo->maxval = isdigit(*buf)?atoi(buf):0;
    CHECK_FOR_ERROR(((pnminfo->maxval<=0) || (pnminfo->maxval>255 && !pnminfo->asciibody)), pnminfo->jmpbuf, BIL_INVALID_OPERATION);
    }

  /* create a new image of the proper size */
  if (!bilResizeImage(context->parent.image, pnminfo->xres, pnminfo->yres))
    longjmp(pnminfo->jmpbuf,1);
  if (!bilClearImage(context->parent.image))
    longjmp(pnminfo->jmpbuf,1);

  /* do it */
  pnminfo->loader(scan, pnminfo, context->parent.image);

  /* free memory */
  pnmscanner_destroy(scan);
  free(pnminfo);

  return BIL_TRUE;
  }


static void pnm_load_ascii(PNMScanner *scan, PNMInfo *info, BILimage image)
  {
  BILubyte *data, *d;
  BILint    x, y, b;
  BILsizei  np;
  BILchar   buf[BUFLEN];
  BILformat format;
  BILtype   type;

  np   = (info->np)?(info->np):1;
  data = malloc(info->xres * np);
  CHECK_FOR_ERROR((!data), info->jmpbuf, BIL_OUT_OF_MEMORY);

  format = (info->np >= 3) ? BIL_RGB : BIL_LUMINANCE;
  type   = BIL_UNSIGNED_BYTE;

  for (y = 0; y < info->yres; y++)
    {
    d = data;

    for (x = 0; x < info->xres; x++)
      {
      for (b = 0; b < np; b++)
        {
        /* Truncated files will just have all 0's at the end of the images */
        CHECK_FOR_ERROR(pnmscanner_eof(scan), info->jmpbuf, BIL_FILE_READ_ERROR);

        if (info->np)
          pnmscanner_gettoken(scan, buf, BUFLEN);
        else
          pnmscanner_getsmalltoken(scan, buf);

        switch (info->maxval)
          {
          case 255:
            d[b] = isdigit(*buf)?atoi(buf):0;
            break;

          case 1:
            d[b] = (*buf=='0')?0xff:0x00;
            break;

          default:
            d[b] = (BILubyte)(255.0*(((BILdouble)(isdigit(*buf)?atoi(buf):0)) / (BILdouble)(info->maxval)));
          }
        }

      d += np;
      }

    bilDrawPixels(image, 0, y, info->xres, 1, format, type, data);
    }

  free(data);
  }


static void pnm_load_raw(PNMScanner *scan, PNMInfo *info, BILimage image)
  {
  FILE     *file;
  BILubyte *data;
  BILint    x, y;
  BILformat format;
  BILtype   type;

  data = malloc(info->xres * info->np);
  CHECK_FOR_ERROR((!data), info->jmpbuf, BIL_OUT_OF_MEMORY);
  
  file   = pnmscanner_file(scan);
  format = (info->np >= 3) ? BIL_RGB : BIL_LUMINANCE;
  type   = BIL_UNSIGNED_BYTE;

  for (y = 0; y < info->yres; y++)
    {
    CHECK_FOR_ERROR((info->xres*info->np != (BILsizei)fread(data, 1,info->xres*info->np, file)), info->jmpbuf, BIL_FILE_READ_ERROR);

    if (info->maxval != 255)	/* Normalize if needed */
      {
      for (x = 0; x < info->xres * info->np; x++)
        data[x] = (BILubyte)(255.0*(BILdouble)(data[x]) / (BILdouble)(info->maxval));
      }

    bilDrawPixels(image, 0, y, info->xres, 1, format, type, data);
    }

  free(data);
  }


static void pnm_load_rawpbm(PNMScanner *scan, PNMInfo *info, BILimage image)
  {
  BILubyte *buf;
  BILubyte  curbyte;
  BILubyte *data;
  BILint    x, y;
  FILE     *file;
  BILsizei  rowlen;
  BILint    bufpos;
  BILformat format;
  BILtype   type;

  format = BIL_LUMINANCE;
  type   = BIL_UNSIGNED_BYTE;

  file   = pnmscanner_file(scan);
  rowlen = (BILint)ceil((BILdouble)(info->xres)/8.0);

  data   = malloc(info->xres);
  CHECK_FOR_ERROR((!data), info->jmpbuf, BIL_OUT_OF_MEMORY);

  buf    = malloc(rowlen*sizeof(BILubyte));
  CHECK_FOR_ERROR((!buf), info->jmpbuf, BIL_OUT_OF_MEMORY);

  for (y = 0; y < info->yres; y++)
    {
    CHECK_FOR_ERROR((rowlen != (BILsizei)fread(buf, 1, rowlen, file)), info->jmpbuf, BIL_FILE_READ_ERROR);

    bufpos  = 0;
    curbyte = buf[0];

    for (x = 0; x < info->xres; x++)
      {
      if ((x % 8) == 0)
        curbyte = buf[bufpos++];

      data[x] = (curbyte&0x80) ? 0x00 : 0xff;
      
      curbyte <<= 1;
      }

    bilDrawPixels(image, 0, y, info->xres, 1, format, type, data);
    }

  free(buf);
  free(data);
  }


/* Writes out RGB and greyscale raw rows */
static void pnmsaverow_raw(PNMRowInfo *ri, BILubyte *data)
  {
  fwrite(data, 1, ri->xres*ri->np, ri->file);
  }


/* Writes out RGB and greyscale ascii rows */
static void pnmsaverow_ascii (PNMRowInfo *ri, BILubyte *data)
  {
  BILint i;  

  for (i = 0; i < ri->xres*ri->np; i++)
    fprintf(ri->file, "%d\n", 0xff & *(data++));
  }


BILboolean ppm_save_image(BILPPMencodeContextStruct *context)
  {  
  PNMRowInfo        rowinfo;
  void           (*saverow)(PNMRowInfo *, BILubyte *) = NULL;
  BILubyte         *data;
  BILsizei          np;
  BILsizei          xres, yres;
  BILint            ypos;
  const BILchar    *comment;
  const BILboolean *writeAsciiPtr;

  np   = 3; /* always RGB data */
  xres = bilGetImageWidth(context->parent.image);
  yres = bilGetImageHeight(context->parent.image);  

  /* allocate a buffer for retrieving information from the pixel region  */
  data = malloc(xres * np);
  if (!data)
    {
    bilSetError(BIL_OUT_OF_MEMORY);
    return BIL_FALSE;
    }

  /* write out magic number */
  writeAsciiPtr = bilGetImageParasiteBoolean(context->parent.image, "ppm_ascii");
  if (writeAsciiPtr && *writeAsciiPtr)
    {
    fwrite("P3\n", 1, 3, context->tempFile);
    saverow = pnmsaverow_ascii;
    }
  else
    {
    fwrite("P6\n", 1, 3,context->tempFile);
    saverow = pnmsaverow_raw;
    }

  /* write out comment string */
  comment = bilGetImageParasiteCharPtr(context->parent.image, "comment");
  if (comment)
    {
    //Handle new-lines in comment!!
    //write (fd, SAVE_COMMENT_STRING, strlen (SAVE_COMMENT_STRING));
    }

  /* write out resolution and maxval */
  fprintf(context->tempFile, "%d %d\n255\n", xres, yres);

  /* set row info */
  rowinfo.file = context->tempFile;
  rowinfo.xres = xres;
  rowinfo.np   = np;

  /* Write the body out */
  for (ypos = 0; ypos < yres; ypos++)
    {
    bilReadPixels(context->parent.image, 0, ypos, xres, 1, BIL_RGB, BIL_UNSIGNED_BYTE, data);
    saverow(&rowinfo, data);
    }

  free(data);

  return BIL_TRUE;
  }


/**************** FILE SCANNER UTILITIES **************/

/* pnmscanner_create ---
 *    Creates a new scanner based on a file.
 */
static PNMScanner *pnmscanner_create(FILE *file)
  {
  PNMScanner *s;

  s = (PNMScanner *)malloc(sizeof(PNMScanner));
  if (s)
    {
    s->file = file;
    s->cur  = ' ';
    }

  return s;
  }


/* pnmscanner_destroy ---
 *    Destroys a scanner and its resources.  Doesn't close the fd.
 */
static void pnmscanner_destroy (PNMScanner *s)
  {
  if (s) free(s);
  }


/* pnmscanner_gettoken ---
 *    Gets the next token, eating any leading whitespace.
 */
static void pnmscanner_gettoken(PNMScanner *s, BILchar *buf, BILsizei bufsize)
  {
  int ctr = 0;

  pnmscanner_eatwhitespace(s);

  while (!pnmscanner_eof(s) && !isspace(s->cur) && (s->cur != '#') && (ctr<bufsize))
    {
    buf[ctr++] = s->cur;
    pnmscanner_getchar(s);
    }
  
  buf[ctr] = '\0';
  }


/* pnmscanner_getsmalltoken ---
 *    Gets the next char, eating any leading whitespace.
 */
static void pnmscanner_getsmalltoken(PNMScanner *s, BILchar *buf)
  {
  pnmscanner_eatwhitespace(s);

  if (!pnmscanner_eof(s) && !isspace(s->cur) && (s->cur != '#'))
    {
    *buf = s->cur;
    pnmscanner_getchar(s);
    }
  }


/* pnmscanner_getchar ---
 *    Reads a character from the input stream
 */
static void pnmscanner_getchar(PNMScanner *s)
  {
  BILint ch;

  ch = fgetc(pnmscanner_file(s));
  if ((ch < 0) || (ch > 127)) ch = ' ';

  s->cur = (BILchar)ch;
  }


/* pnmscanner_eatwhitespace ---
 *    Eats up whitespace from the input and returns when done or eof.
 *    Also deals with comments.
 */
static void pnmscanner_eatwhitespace(PNMScanner *s)
  {
  int state = 0;

  while (!pnmscanner_eof(s) && (state != -1))
    {
    switch (state)
      {
      case 0:  /* in whitespace */
        if (s->cur == '#')
          {
          state = 1;  /* goto comment */
          pnmscanner_getchar(s);
          }
        else if (!isspace(s->cur))
          state = -1;
        else
          pnmscanner_getchar(s);
        break;

      case 1:  /* in comment */
        if (s->cur == '\n')
          state = 0;  /* goto whitespace */
        pnmscanner_getchar(s);
        break;
      }
    }
  }

